use std::fs::{remove_dir, remove_dir_all};
use std::process::exit;

use log::{debug, error, info, trace};

use errors::{BinError, BuildError, InstallError};
use manifest::Manifest;
use pkgfile::PKGFile;

use crate::CONFIG;
use crate::package::identifier::{PackageIdentifier, PackageLocator};
use crate::package::Package;
use crate::tmpfs::TempDir;
use crate::types::fetch::TryFetch;
use crate::util::{visit_dirs};
use crate::util::prompts::prompt_bool;

pub mod errors;

#[derive(Debug)]
pub enum InstallType {
    Build,
    Bin,
}

#[derive(Debug)]
pub struct PackageInstaller {
    manifest: Manifest,
    pkgfile: PKGFile,
    install_type: InstallType,
}

impl PackageInstaller {
    pub fn new(m: Manifest, p: PKGFile, i: InstallType) -> PackageInstaller {
        PackageInstaller {
            manifest: m,
            pkgfile: p,
            install_type: i,
        }
    }

    fn extract_to(&self, path: String) -> Result<(), BinError> {
        if !self.pkgfile.has_data() {
            return Err(BinError::UnpackError("package has no data".to_string()));
        }
        let path = std::path::Path::new(&path);
        if path.exists() {
            trace!("cache already exists..");
            debug!("removing cache dir...");
            match remove_dir_all(path) {
                Ok(_) => debug!("removed cache directory"),
                Err(e) => return Err(BinError::UnpackError(format!("unable to remove directory ({}): {}", path.to_str().unwrap_or("unknown"), e.to_string())))
            }
        }
        tar::Archive::new(self.pkgfile.data.as_slice())
            .unpack(&path)
            .map_err(|e| BinError::UnpackError(e.to_string()))
    }

    fn bin(&self) -> Result<(), BinError> {
        let mut tmpdir = TempDir::default();
        tmpdir.push(&self.manifest.package.name);
        trace!("extracting package into: {}", tmpdir.to_string());
        if let Err(e) = self.extract_to(tmpdir.to_string()) {
            return Err(e)
        }
        debug!("extracted package in: {}", tmpdir.to_string());
        if prompt_bool("See all files?", false) {
            if let Err(e) = visit_dirs(tmpdir.path().as_path(), &|d| info!(target: "item", "{}", d.path().to_str().unwrap())) {
                error!("Could not show files: {}", e);
            }
            if prompt_bool("Continue?", false) {
                self.apply_overlay();
            }
        } else {
            self.apply_overlay();
        }
        Ok(())
    }

    fn apply_overlay(&self) {

    }

    fn build(&self) -> Result<(), BuildError> {
        if let None = self.manifest.build.clone() {
            return Err(BuildError::InvalidManifest);
        }
        let build_manifest = self.manifest.build.clone().unwrap();
        // TODO: Check dependencies
        for pkg_tuple in build_manifest.dependencies {
            let mut pkg = Package::try_fetch(
                PackageIdentifier::PackageLocator(
                    PackageLocator::from(pkg_tuple)
                )
            ).expect("no pkg");

            if !pkg.is_installed() {
                match pkg.install(CONFIG.with(|c| {
                    c.build_by_default
                })) {
                    Ok(_) => { info!("Installed dependency: \"{}\"", pkg.manifest().package.name) }
                    Err(_) => {
                        error!("Could not install dependency: \"{}\"", pkg.identifier);
                        exit(1);
                    }
                }
            }
        }

        Ok(())
    }

    pub fn install(&mut self) -> Result<(), InstallError> {
        match self.install_type {
            InstallType::Bin => {
                if let None = self.manifest.bin {
                    self.install_type = InstallType::Build;
                    if let None = self.manifest.build {
                        return Err(InstallError::InvalidManifest);
                    }
                }
            }
            InstallType::Build => {
                if let None = self.manifest.build {
                    self.install_type = InstallType::Bin;
                    if let None = self.manifest.bin {
                        return Err(InstallError::InvalidManifest);
                    }
                }
            }
        }

        match self.install_type {
            InstallType::Bin => self.bin()
                .map_err(|e| InstallError::BinError(e)),
            InstallType::Build => self.build()
                .map_err(|e| InstallError::BuildError(e)),
        }
    }
}
